import logging

from nymms.resources import load_resource, Node
from nymms.registry import DuplicateEntryError

logger = logging.getLogger(__name__)


class Backend(object):
    def load_nodes(self):
        """ Should return a dictionary of Node information in this form:
        {'<node_name>': {<node creation kwargs>}, ...}

        Meant to be overridden by subclasses.
        """
        raise NotImplementedError

    def _load_nodes(self, reset=False):
        nodes = self.load_nodes()
        try:
            load_resource(nodes, Node, reset=reset)
        except DuplicateEntryError, e:
            # TODO: Need to figure out the story for reloading nodes, etc.
            logger.debug("Nodes already loaded and reset is False, skipping.")



import time
import logging
import copy

logger = logging.getLogger(__name__)

from nymms.data_types import NymmsDataType, ValidationError

# state constants
OK = 0
WARN = WARNING = 1
CRIT = CRITICAL = 2
UNKNOWN = 3

# Anything over state 2 is unknown
states = ['ok', 'warning', 'critical', 'unknown']

# state type constants
SOFT = 0
HARD = 1

state_types = ['soft', 'hard']


def get_state_type_name(state_type_code):
    return state_types[state_type_code]


def get_state_name(state_code):
    if state_code > 3:
        return "unknown"
    return states[state_code]


def validate_state(state):
    value = state
    exc = ValidationError('state', value)
    if isinstance(value, basestring):
        try:
            return states.index(value.lower())
        except ValueError:
            raise exc
    elif isinstance(value, int):
        if value >= 0:
            return value
    raise exc


def validate_state_type(state_type):
    value = state_type
    exc = ValidationError('state_type', value)
    if isinstance(value, basestring):
        try:
            return state_types.index(value.lower())
        except ValueError:
            raise exc
    elif isinstance(value, int):
        try:
            return state_types[value] and value
        except IndexError:
            raise exc
    raise exc


class StateMixin(object):
    def validate_state(self):
        self.state = validate_state(self.state)

    def validate_state_type(self):
        self.state_type = validate_state_type(self.state_type)

    @property
    def state_name(self):
        self.validate_state()
        return get_state_name(self.state)

    @property
    def state_type_name(self):
        self.validate_state_type()
        return get_state_type_name(self.state_type)


class Result(NymmsDataType, StateMixin):
    required_fields = ['state', 'state_type']

    def __init__(self, object_id, state=None, state_type=None, timestamp=None,
                 output=None, task_context=None, origin=None):
        super(Result, self).__init__(object_id=object_id, origin=origin)
        self.state = state
        self.state_type = state_type
        self.timestamp = timestamp or time.time()
        self.output = output or ''
        self.task_context = task_context or {}

    def validate_timestamp(self):
        self.timestamp = int(self.timestamp)

    def _serialize(self):
        self._cleaned['state_name'] = get_state_name(self.state)
        self._cleaned['state_type_name'] = get_state_type_name(self.state_type)

    @classmethod
    def _deserialize(cls, item):
        new_item = super(Result, cls)._deserialize(item)
        new_item = copy.deepcopy(item)
        del(new_item['state_name'])
        del(new_item['state_type_name'])
        return new_item


class StateRecord(NymmsDataType, StateMixin):
    required_fields = ['state', 'state_type']

    def __init__(self, object_id, last_update=None, last_state_change=None,
                 state=None, state_type=None, origin=None):
        super(StateRecord, self).__init__(object_id=object_id, origin=origin)
        self.last_update = last_update
        self.last_state_change = last_state_change
        self.state = state
        self.state_type = state_type
        self._cleaned = {}

    def validate_last_update(self):
        self.last_update = int(self.last_update or time.time())

    def validate_last_state_change(self):
        self.last_state_change = int(self.last_state_change or time.time())

    @classmethod
    def decode_value(cls, value):
        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                return str(value)

    @classmethod
    def _deserialize(cls, item):
        item_dict = {}
        for k, v in item.iteritems():
            item_dict[k] = cls.decode_value(v)
        return item_dict



import logging
import copy
import os
import collections

logger = logging.getLogger(__name__)

from nymms.config import yaml_config


def deep_update(orig, upd):
    """ Does a 'deep' update of dictionary 'orig' with dictionary 'upd'."""
    for k, v in upd.iteritems():
        if isinstance(v, collections.Mapping):
            r = deep_update(orig.get(k, {}), v)
            orig[k] = r
        else:
            orig[k] = upd[k]
    return orig 

default_conf_dir = '/etc/nymms'

DEFAULTS = {
    'monitor_timeout': 15,
    'resources': os.path.join(default_conf_dir, 'resources.yaml'),
    'region': 'us-east-1',
    'state_domain': 'nymms_state',
    'tasks_queue': 'nymms_tasks',
    'results_topic': 'nymms_results',
    'private_context_file': os.path.join(default_conf_dir, 'private.yaml'),
    'task_expiration': 600,

    'probe': {
        'max_retries': 2,
        'queue_wait_time': 20,
        'retry_delay': 30,
        'blah': 22,
    },

    'reactor': {
        'handler_config_path': os.path.join(default_conf_dir, 'handlers'),
        'queue_name': 'reactor_queue',
        'queue_wait_time': 20,
        'visibility_timeout': 30,
    },

    'scheduler': {
        'interval': 300,
        'backend': 'nymms.scheduler.backends.yaml_backend.YamlBackend',
        'backend_args': {
            'path': os.path.join(default_conf_dir, 'nodes.yaml'),
        }
    },
}

settings = None
version = None


def load_config(path, force=False):
    global settings, version, DEFAULTS
    if settings and not force:
        return
    settings = copy.deepcopy(DEFAULTS)
    version, _config_settings = yaml_config.load_config(path)
    if _config_settings:
        deep_update(settings, _config_settings)
    logger.debug("Config loaded from '%s' with version '%s'." % (path,
                                                                 version))



import logging
import copy

from nymms.exceptions import NymmsException

logger = logging.getLogger(__name__)


class ValidationError(NymmsException):
    def __init__(self, field, value):
        self.field = field
        self.value = value

    def __str__(self):
        return "Invalid value (%s) in field: %s" % (self.value, self.field,)


class MissingRequiredField(NymmsException):
    def __init__(self, field):
        self.field = field

    def __str__(self):
        return "Required field is blank: %s" % (self.field,)


class NymmsDataType(object):
    required_fields = []

    def __init__(self, object_id, origin=None):
        self.id = object_id
        self._origin = origin
        self._cleaned = {}

    def __str__(self):
        return "%s: %s" % (self.__class__.__name__, self.id)

    def validate(self):
        for field in self.required_fields:
            if getattr(self, field, None) is None:
                raise MissingRequiredField(field)
        for attr in dir(self):
            if attr.startswith('validate_'):
                validate_method = getattr(self, attr)
                if callable(validate_method):
                    validate_method()

    def delete(self):
        if self._origin:
            return self._origin.delete()

    def _serialize(self):
        """ Children classes should override this to update the dictionary
        returned by the serialize method.
        """
        return True

    def serialize(self, force=False):
        if self._cleaned and not force:
            return self._cleaned

        self.validate()
        for k, v in self.__dict__.iteritems():
            if not k.startswith('_'):
                self._cleaned[k] = v
        self._serialize()
        return self._cleaned

    @staticmethod
    def _deserialize(data):
        """ Children classes should override this to modify the dictionary
        that is used by the deserialize method.
        """
        return copy.deepcopy(data)

    @classmethod
    def deserialize(cls, data, origin=None):
        """ Produces a new data type object from a dictionary of data.  The
        origin object is passed in for further interactions with the system
        that the object was received from.  For example to delete the object
        out of a queue.
        """
        new_data = cls._deserialize(data)
        if not origin:
            origin = data
        object_id = new_data.pop('id')
        return cls(object_id, origin=origin, **new_data)



import logging

from nymms.reactor.handlers.Handler import Handler
from nymms.utils.aws_helper import ConnectionManager
from nymms import results

from jinja2 import Template

logger = logging.getLogger(__name__)


class SESHandler(Handler):
    """ A basic handler to send alerts to people via email through Amazon's
    SES service.  Sends every result it receives by default.  To filter
    results you should subclass this and provide a _filter method.

    config options:
      enabled: bool
      region: string, aws region (us-east-1, etc)
      sender: string, email address
      subject_template: string
      body_template: string
      recipients: list, email addresses
      filters: list, filters
    """
    def _connect(self):
        if getattr(self, '_aws_conn', None):
            return
        self._aws_conn = ConnectionManager(region=self.config['region'])

    def _send_email(self, result, previous_state):
        self._connect()
        subject = Template(self.config['subject_template'])
        body = Template(self.config['body_template'])
        sender = self.config['sender']
        recipients = self.config['recipients']
        result_data = result.serialize()
        logger.debug("Sending SES alert to %s as %s for %s.",
                     recipients, sender, result.id)
        self._aws_conn.ses.send_email(
            source=sender,
            subject=subject.render(result_data),
            body=body.render(result_data),
            to_addresses=recipients
        )

    def process(self, result, previous_state):
        self._send_email(result, previous_state)



import logging

logger = logging.getLogger(__name__)

import nymms

class NymmsDaemon(object):
    def __init__(self):
        logger.debug("%s initialized.", self.__class__.__name__)

    def main(self, *args, **kwargs):
        logger.debug("Launching %s version %s.", self.__class__.__name__,
                     nymms.__version__)
        self.run(*args, **kwargs)
                     



import logging
import json

logger = logging.getLogger(__name__)

from boto.sqs.message import Message

from nymms.tasks import Task
from nymms.probe.Probe import Probe
from nymms.state.sdb_state import SDBStateBackend
from nymms.utils.aws_helper import SNSTopic


class SQSProbe(Probe):
    def __init__(self, conn_mgr, task_queue, results_topic, state_domain,
                 state_backend=SDBStateBackend):
        self._conn = conn_mgr
        self._queue_name = task_queue
        self._topic_name = results_topic
        self._topic = None
        self._queue = None
        self._state_backend = state_backend(conn_mgr.sdb, state_domain)
        super(SQSProbe, self).__init__()

    def _setup_queue(self, **kwargs):
        if self._queue:
            return
        logger.debug("setting up queue %s", self._queue_name)
        self._queue = self._conn.sqs.create_queue(self._queue_name)

    def _setup_topic(self, **kwargs):
        if self._topic:
            return
        logger.debug("setting up topic %s", self._topic_name)
        self._topic = SNSTopic(self._conn, self._topic_name)

    def get_task(self, **kwargs):
        self._setup_queue(**kwargs)
        wait_time = kwargs.get('queue_wait_time')
        timeout = kwargs.get('monitor_timeout') + 3
        logger.debug("Getting task from queue %s.", self._queue_name)
        task_item = self._queue.read(visibility_timeout=timeout,
                                     wait_time_seconds=wait_time)
        task = None
        if task_item:
            task = Task.deserialize(json.loads(task_item.get_body()),
                                    origin=task_item)
        return task

    def resubmit_task(self, task, delay, **kwargs):
        self._setup_queue(**kwargs)
        task.increment_attempt()
        logger.debug("Resubmitting task %s with %d second delay.", task.id,
                     delay)
        m = Message()
        m.set_body(json.dumps(task.serialize()))
        return self._queue.write(m, delay_seconds=delay)

    def submit_result(self, result, **kwargs):
        self._setup_topic()
        logger.debug("%s - submitting '%s/%s' result", result.id,
                     result.state_name, result.state_type_name)
        return self._topic.publish(json.dumps(result.serialize()))



import logging
import glob
import os
import sys

import nymms
from nymms.daemon import NymmsDaemon
from nymms.config import yaml_config
from nymms.utils import load_object_from_string, logutil
from nymms.exceptions import OutOfDateState

logger = logging.getLogger(__name__)


class Reactor(NymmsDaemon):
    def __init__(self):
        self._handlers = {}
        super(Reactor, self).__init__()

    def _load_handlers(self, handler_config_path, **kwargs):
        base_path = os.path.expanduser(handler_config_path)
        conf_files = glob.glob(os.path.join(base_path, '*.conf'))
        logger.debug("Loading handlers from %s", handler_config_path)
        for f in conf_files:
            handler_name = os.path.split(f)[1][:-5]
            # We could eventually have the handlers get loaded everytime and
            # update them if their config has changed (via config_version
            # below).  For now lets not get that tricky.
            if handler_name in self._handlers:
                if not kwargs.get('force_load_handlers', False):
                    logger.debug("Handler %s already loaded, skipping.",
                                 handler_name)
                    continue
            conf_version, conf = yaml_config.load_config(f)
            enabled = conf.pop('enabled', False)
            if not enabled:
                logger.debug("Handler %s 'enabled' is not set to true. "
                             "Skipping.", handler_name)
                continue
            cls_string = conf.pop('handler_class')
            logger.debug('Initializing handler %s.', handler_name)
            try:
                handler_cls = load_object_from_string(cls_string)
                self._handlers[handler_name] = handler_cls(conf)
            except Exception as e:
                logutil.log_exception("Skipping handler %s due to "
                    "unhandled exception:" % (handler_name), logger)
                continue

        if not self._handlers:
            logger.error("No handlers loaded.  Exiting.")
            sys.exit(1)

    def get_result(self, **kwargs):
        raise NotImplementedError

    # TODO: This calls on _state_backend but setting up of the _state_backend
    #       needs to be handled in the subclass.  Not sure how I should handle
    #       this, but I really like the idea of these being base class
    #       methods since in reality all reactors should have some sort of
    #       state backend, even if its a no-op
    def get_state(self, task_id):
        return self._state_backend.get_state(task_id)

    def save_state(self, task_id, result, previous):
        return self._state_backend.save_state(task_id, result, previous)

    def handle_result(self, result, **kwargs):
        previous_state = self.get_state(result.id)
        for handler_name, handler in self._handlers.iteritems():
            try:
                handler._process(result, previous_state)
            except Exception:
                logutil.log_exception("Unhandled %s handler "
                                      "exception:" % (handler_name,), logger)
                continue
        try:
            self.save_state(result.id, result, previous_state)
        except OutOfDateState:
            pass

    def run(self, handler_config_path, **kwargs):
        """ This will run in a tight loop. It is expected that the subclass's
        get_result() method will introduce a delay if the results queue is
        empty.
        """
        self._load_handlers(handler_config_path, **kwargs)
        while True:
            result = self.get_result(**kwargs)
            if not result:
                logger.debug('Result queue empty.')
                continue
            self.handle_result(result, **kwargs)
            result.delete()



import logging

logger = logging.getLogger(__name__)

from nymms.utils import load_object_from_string


class Handler(object):
    def __init__(self, config=None):
        self.config = config
        self._filters = []

    def _load_filters(self):
        filters = self.config.get('filters', [])
        if filters:
            for filter_string in filters:
                logging.debug("Adding Filter %s to Handler %s.", filter_string,
                              self.__class__.__name__)
                f = load_object_from_string(filter_string)
                self._filters.append(f)
        else:
            logger.debug("No filters configured for Handler %s.",
                         self.__class__.__name__)

    def _filter(self, result, previous_state):
        """ Runs the result & previous state through all the configured
        filters.  A filter should be a callable that accepts two arguments:
        the result and the previous state.  It should return either True or
        False regarding whether the message should be allowed through the
        handler.
        """
        if not self._filters:
            self._load_filters()
        # Assume that no filters means just that - that the result is
        # not to be filtered for the handler.
        results = [True]
        for f in self._filters:
            try:
                results.append(f(result, previous_state))
            except Exception as e:
                logger.exception("Filter %s on Handler %s had an unhandled "
                                 "exception. Ignoring:",
                                 f.__name__, self.__class__.__name__)
                continue
        return all(results)

    def _process(self, result, previous_state):
        if self._filter(result, previous_state):
            logger.debug("Handler %s filters returned true for %s, reacting.",
                         self.__class__.__name__, result.id)
            return self.process(result, previous_state)
        logger.debug("Handler %s filters returned false for %s, skipping.",
                     self.__class__.__name__, result.id)

    def process(self, result, previous_state):
        """ Meant to be overridden by subclasses - should handle the actual
        process of reacting to a result.
        """
        raise NotImplementedError



import logging
from logging.handlers import SysLogHandler
import os.path
import sys
import traceback
import platform
import os

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

_pid = os.getpid()

syslog_socket_paths = {
    'Darwin': '/var/run/syslog',
    'Linux': '/dev/log'
}


def quiet_boto_logging():
    """
    Boto's debug logs are full dumps of the XML that was passed between the
    client and server.  This can be annoying.  This is a simple function to
    hide those dumps whenever you put your code into debug.
    """
    logging.getLogger('boto').setLevel(logging.CRITICAL)


def quiet_paramiko_logging():
    """ Paramiko is really noisy when set to INFO or below.

    This sets the paramiko logger to only send WARNING or above messages.
    """
    logging.getLogger('paramiko').setLevel(logging.WARNING)


def quiet_requests_connpool_logging():
    """ Paramiko is really noisy when set to INFO or below.

    This sets the paramiko logger to only send WARNING or above messages.
    """
    logging.getLogger('requests.packages.urllib3.connectionpool').setLevel(
        logging.WARNING)


def get_syslog_path():
    system_os = platform.system()
    try:
        return syslog_socket_paths[system_os]
    except KeyError:
        raise ValueError("Unable to find syslog unix domain socket for os "
                         "'%s'." % (system_os))

DEFAULT_FORMAT = ('pid:' + str(_pid) + ' %(levelname)s %(name)s '
                  '%(module)s(%(funcName)s):%(lineno)d - %(message)s')


def setup_root_logger(stdout=INFO, filename=None, file_level=INFO,
                      file_mode='w', syslog=None,
                      syslog_facility=SysLogHandler.LOG_LOCAL7,
                      syslog_socket_path=None, syslog_tag=None,
                      time_format="%Y/%m/%d %H:%M:%S %Z",
                      message_format=DEFAULT_FORMAT):
    """Setup basic logging, including stdout, file, and syslog logging.

    Sets up the root logger, deleting any previously configured handlers.  It
    does this to make sure that we don't have multiples of the same handler
    being attached to the root logger, resulting in multiple messages of the
    same type.

    This should be called in the main script/command/daemon itself, and never
    inside libraries unless you really know what you're doing.

    :type stdout: int
    :param stdout: The logging level to send to stdout.  Can be any of the
        logging.* constants (logging.DEBUG, etc) or logutil constants
        (logutil.DEBUG, etc) which are just pointers to the logging constants.
        If set to None or False, disable stdout logging.
        Default: logging.INFO

    :type filename: string
    :param filename: The path to a file to log to.  Setting to None disables
        file logging.
        Default: None

    :type file_level: int
    :param file_level: The logging level to send to the file given in the
        'filename' parameter.  Can be any of the logging.* or logutil.*
        constants like the 'stdout' parameter.
        Default: logging.INFO

    :type file_mode: string
    :param file_mode: The mode to open the file at the 'filename' parameter
        with.
        Default: 'w'

    :type syslog: int
    :param syslog: The logging level to send to syslog.  Can be any of the
        logging.* or logutil.* constants.  Set to None to disable syslog
        logging.
        Default: None

    :type syslog_facility: int
    :param syslog_facility: The syslog facility to send messages to if syslog
        is enabled.  Can be any of the SysLogHandler.LOG_* facility constants.
        Default: SysLogHandler.LOG_LOCAL7

    :type syslog_socket_path: string
    :param syslog_socket_path: The path to the unix domain socket used by
        syslog if syslog is enabled.  If not given, will automatically try to
        determine the correct path.
        Default: None

    :type syslog_tag: string
    :param syslog_tag: The tag to be pre-pended to syslog messages.  If not
        given it will try to determine the name of the command that was
        called, and use that.
        Default: None

    :type time_format: string
    :param time_format: A time.strftime formatted string to use for the
        timestamp format.  This will be prepended to stdout and logfiles, but
        not to syslog (since syslog has it's own timestamp system)

    :type message_format: string
    :param message_format: A logging.Formatter formatted string to use for
        the output of log messages.  See the following for variables:
        http://docs.python.org/2/library/logging.html#logrecord-attributes
    """
    base_format = message_format
    timed_format = '[%(asctime)s] ' + base_format
    timed_formatter = logging.Formatter(timed_format, datefmt=time_format)
    logger = logging.getLogger()

    # Delete all previous handlers.
    for h in logger.handlers:
        logger.removeHandler(h)

    # Used to track what levels are being used by handlers.
    levels = []

    if stdout:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(timed_formatter)
        stdout_handler.setLevel(stdout)
        levels.append(stdout)
        logger.addHandler(stdout_handler)

    if filename:
        file_handler = logging.FileHandler(filename, mode=file_mode)
        file_handler.setFormatter(timed_formatter)
        file_handler.setLevel(file_level)
        levels.append(file_level)
        logger.addHandler(file_handler)

    if syslog:
        if not syslog_socket_path:
            syslog_socket_path = get_syslog_path()
        syslog_handler = SysLogHandler(syslog_socket_path,
                                       facility=syslog_facility)
        if not syslog_tag:
            syslog_tag = os.path.basename(sys.argv[0])
        syslog_format = syslog_tag + ": " + base_format
        syslog_handler.setFormatter(logging.Formatter(syslog_format))
        syslog_handler.setLevel(syslog)
        levels.append(syslog)
        logger.addHandler(syslog_handler)

    # Set the logger level to the level of the lowest leveled handler
    logger.setLevel(min(levels))

    return logger


def log_exception(message=None, logger=logging):
    """
    Used to produce more cleanly readable exceptions in syslog by breaking
    the exception up over multiple logging calls.
    """
    if message:
        logger.error(message)
    logger.error('Exception output: ')
    exc_msg = traceback.format_exc().split('\n')
    for line in exc_msg:
        logger.error('    %s' % (line,))



import logging

logger = logging.getLogger(__name__)

import copy
from weakref import WeakValueDictionary

from nymms import registry
from nymms.utils import commands
from nymms.config import yaml_config
from nymms.exceptions import MissingCommandContext

from jinja2 import Template
from jinja2.runtime import StrictUndefined
from jinja2.exceptions import UndefinedError


RESERVED_ATTRIBUTES = ['name', 'address', 'node_monitor', 'monitoring_groups',
                       'command_string']


class RegistryMetaClass(type):
    """ Creates a registry of all objects of a classes type.

    This allows us to get a list of all objects of a given class type quickly
    and easily.  IE:

    >>> from nymms import resources
    >>> webservers = resources.MonitoringGroup('webservers')
    >>> www1 = resources.Node('www1', monitoring_groups=[webservers])
    >>> resources.Node.registry.items()
    [('www1', <nymms.resources.Node object at 0x10b6aa8d0>)]
    """
    def __new__(cls, name, bases, dct):
        new_class = super(RegistryMetaClass, cls).__new__(cls, name, bases,
                                                          dct)
        new_class.registry = registry.Registry(new_class)
        return new_class


class NanoResource(object):
    __metaclass__ = RegistryMetaClass

    context_attributes = ['name']

    def __init__(self, name, **kwargs):
        self.name = name
        # Ensure noone tries to set a reserved attribute as an extra
        disallowed_attributes = list(
            set(RESERVED_ATTRIBUTES) & set(kwargs.keys()))
        if disallowed_attributes:
            raise TypeError("The following are reserved attributes and cannot "
                            "be used on this resource: %s" % (', '.join(
                                disallowed_attributes)))
        self.extra_attributes = kwargs
        self._context_cache = None

        self.register()

    def register(self):
        logger.debug("Registering %s resource '%s' with the '%s' registry." % (
            self.__class__.__name__, self.name, self.__class__.__name__))
        self.registry[self.name] = self

    def _context(self, force=False):
        if self._context_cache and not force:
            logger.debug("Returning context cache for %s resource." % (
                self.name))
            return self._context_cache
        logger.debug("Generating context cache for %s resource." % (self.name))
        context_key = self.__class__.__name__.lower()
        context = {}
        for attr in self.context_attributes:
            context[attr] = getattr(self, attr)
        for k, v in self.extra_attributes.items():
            context[k] = v
        self._context_cache = {context_key: context}
        return self._context_cache


class MonitoringGroup(NanoResource):
    context_attributes = ['name', 'realm']

    def __init__(self, name, realm=None, monitors=None, nodes=None, **kwargs):
        self.nodes = WeakValueDictionary()
        self.monitors = WeakValueDictionary()
        self.realm = realm

        super(MonitoringGroup, self).__init__(name, **kwargs)

        if monitors:
            for monitor in monitors:
                self.add_monitor(monitor)
        if nodes:
            for node in nodes:
                self.add_node(node)

    def add_node(self, node):
        if not isinstance(node, Node):
            try:
                node = Node.registry[node]
            except KeyError:
                logger.error("Unable to find Node '%s' in registry." % (
                    node))
        logger.debug("Adding node '%s' to monitoring group '%s'." % (node.name,
                     self.name))
        self.nodes[node.name] = node
        node.monitoring_groups[self.name] = self

    def add_monitor(self, monitor):
        if not isinstance(monitor, Monitor):
            try:
                monitor = Monitor.registry[monitor]
            except KeyError:
                logger.error("Unable to find Monitor '%s' in registry." % (
                    monitor))
        logger.debug("Adding monitor '%s' to monitoring group '%s'." % (
            monitor.name, self.name))
        self.monitors[monitor.name] = monitor
        monitor.monitoring_groups[self.name] = self


class Node(NanoResource):
    context_attributes = ['name', 'realm', 'address', 'node_monitor']

    def __init__(self, name, realm=None, address=None, node_monitor=None,
                 monitoring_groups=None, **kwargs):
        self.name = name
        self.realm = realm
        self.address = address or name
        self.node_monitor = node_monitor
        self.monitoring_groups = WeakValueDictionary()
        self._tasks = []
        if monitoring_groups:
            for group in monitoring_groups:
                if not isinstance(group, MonitoringGroup):
                    try:
                        group = MonitoringGroup.registry[group]
                    except KeyError:
                        logger.error("Unable to find MonitoringGroup '%s' "
                                     "in registry, skipping." % (group))
                group.add_node(self)

        super(Node, self).__init__(name, **kwargs)

    def build_context(self, monitoring_group, monitor):
        context = {}
        for obj in (monitoring_group, self, monitor):
            c = obj._context()
            context.update(c)
            for k, v in c.values()[0].iteritems():
                if not k == 'name':
                    context[k] = v
        return context

    @property
    def monitors(self):
        if self._tasks:
            return self._tasks
        for group_name, group in self.monitoring_groups.iteritems():
            for monitor_name, monitor in group.monitors.iteritems():
                self._tasks.append(self.build_context(group, monitor))
        return self._tasks


class Monitor(NanoResource):
    context_attributes = ['name', 'realm']

    def __init__(self, name, command, realm=None, monitoring_groups=None,
                 **kwargs):
        self.name = name
        self.realm = realm
        if not isinstance(command, Command):
            try:
                command = Command.registry[command]
            except KeyError:
                logger.error("Unable to find Command '%s' in registry." % (
                    command))
                raise

        self.command = command
        self.monitoring_groups = WeakValueDictionary()
        if monitoring_groups:
            for group in monitoring_groups:
                if not isinstance(group, MonitoringGroup):
                    try:
                        group = MonitoringGroup.registry[group]
                    except KeyError:
                        logger.error("Unable to find MonitoringGroup '%s' in "
                                     "registry." % (group))
                        raise
                group.add_monitor(self)

        super(Monitor, self).__init__(name, **kwargs)

    def execute(self, context, timeout, private_context=None):
        return self.command.execute(context, timeout, private_context)

    def format_command(self, context, private_context=None):
        return self.command.format_command(context, private_context)


class Command(NanoResource):
    context_attributes = ['name', 'command_type', 'command_string']

    def __init__(self, name, command_string, command_type='nagios', **kwargs):
        self.command_type = command_type
        self.command_string = command_string
        super(Command, self).__init__(name, **kwargs)

    def format_command(self, context, private_context=None):
        my_context = self._context()
        local_context = copy.deepcopy(context)
        local_context.update(my_context)
        local_context['__private'] = {}
        if private_context:
            local_context['__private'].update(private_context)
        for k, v in my_context.values()[0].iteritems():
            if not k == 'name' and not k in local_context:
                local_context[k] = v
        t = Template(self.command_string)
        t.environment.undefined = StrictUndefined
        try:
            out = t.render(local_context)
        except UndefinedError as e:
            raise MissingCommandContext(e.message)
        return t.render(local_context)

    def execute(self, context, timeout, private_context=None):
        cmd = self.format_command(context, private_context)
        return commands.execute(cmd, timeout)


def load_resource(resources, resource_class, reset=False):
    """ Given a dictionary of a given resource_type, instantiate them.

    The resources are loaded into the given resource registry.
    """
    if reset:
        logger.debug("Clearing old %s entries from registry." % (
            resource_class.__name__))
        resource_class.registry.clear()

    for name, kwargs in resources.items():
        if not kwargs:
            kwargs = {}
        resource_class(name, **kwargs)


def load_resources(resource_file, reset=False):
    """ Loads resources in yaml formatted resource_file in the proper order.

    Returns a sha512 hash of the resources.  The resources themselves are
    stored in their individual registries.
    """
    LOAD_ORDER = [('commands', Command),
                  ('monitoring_groups', MonitoringGroup),
                  ('monitors', Monitor)]

    logger.info("Loading local resources from %s." % (resource_file))
    version, resources = yaml_config.load_config(resource_file)

    for resource_type, resource_class in LOAD_ORDER:
        items = resources[resource_type]
        load_resource(items, resource_class, reset=reset)
    return version


def load_nodes(node_file, reset=False):
    """ Loads nodes from a yaml formatted file.

    Nodes are stored in the Node registry.
    """
    logger.info("Loading nodes from %s." % (node_file))
    version, nodes = yaml_config.load_config(node_file)

    items = nodes['nodes']
    load_resource(items, Node, reset=reset)
    return version



class NymmsException(Exception):
    """ NYMMS base exception class.
    """
    pass


class OutOfDateState(NymmsException):
    def __init__(self, current, previous):
        self.current_state = current
        self.previous_state = previous

    def __str__(self):
        return "Previous (%d) state newer than current (%d)." % (
            self.previous_state.last_update, self.current_state.last_update)


class MissingCommandContext(NymmsException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "Invalid command variable: %s" % self.message



import unittest
import time

from nymms.probe.Probe import Probe
from nymms import results
from nymms.tasks import Task

result_codes = [
    results.CRITICAL,
    results.WARNING,
    results.CRITICAL,
    results.WARNING,
    results.WARNING,
    results.OK,
    results.OK,
    results.UNKNOWN,
    results.OK,
    results.OK
]


class DummyStateBackend(object):
    def __init__(self):
        self.states = [
            None,
            {'last_update': 1, 'last_state_change': 0,
                'state': results.CRITICAL, 'state_type': results.SOFT},
            {'last_update': 2, 'last_state_change': 0,
                'state': results.WARNING, 'state_type': results.SOFT},
            {'last_update': 3, 'last_state_change': 3,
                'state': results.CRITICAL, 'state_type': results.HARD},
            {'last_update': 4, 'last_state_change': 4,
                'state': results.WARNING, 'state_type': results.HARD},
            {'last_update': 5, 'last_state_change': 4,
                'state': results.WARNING, 'state_type': results.HARD},
            {'last_update': 6, 'last_state_change': 6,
                'state': results.OK, 'state_type': results.HARD},
            {'last_update': 7, 'last_state_change': 6,
                'state': results.OK, 'state_type': results.HARD},
            {'last_update': 8, 'last_state_change': 8,
                'state': results.UNKNOWN, 'state_type': results.SOFT},
            {'last_update': 9, 'last_state_change': 9,
                'state': results.OK, 'state_type': results.SOFT},
            {'last_update': 10, 'last_state_change': 9,
                'state': results.OK, 'state_type': results.HARD}
        ]
        self.state_iter = iter(self.states)

    def get_state(self, task_id):
        state_item = next(self.state_iter)
        state = None
        if state_item:
            state_item['id'] = task_id
            state = results.StateRecord.deserialize(state_item)
            state.validate()
        return state


class DummyProbe(Probe):
    def __init__(self):
        self.task = Task('test:task', context={})
        self.results_iter = iter(result_codes)
        
    def get_task(self, **kwargs):
        return self.task

    def resubmit_task(self, task, delay, **kwargs):
        self.task.increment_attempt()

    def submit_result(self, result, **kwargs):
        if result.state_type == results.HARD:
            self.task.attempt = 0
        return result

    def execute_task(self, task, timeout, **kwargs):
        result = results.Result(task.id, timestamp=task.created,
                                task_context=task.context)
        result.state = next(self.results_iter)
        result.output = 'Some output here.'
        return result


class TestStateChange(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.state_backend = DummyStateBackend()
        cls.probe = DummyProbe()
        cls.probe._state_backend = cls.state_backend

    def test_state_change(self):
        # tests that our logic follows the nagios logic here
        # http://nagios.sourceforge.net/docs/3_0/statetypes.html
        # We take our example state changes from the table at the bottom
        t = self.probe.get_task()
        for i, code in enumerate(result_codes):
            r = self.probe.handle_task(t, monitor_timeout=30,
                                       max_retries=2)
            expected = self.state_backend.states[i + 1]
            print "[%d] Result STATE/TYPE: %s/%s" % (i, r.state, r.state_type)
            print "[%d] Expected STATE/TYPE: %s/%s" % (i, expected['state'],
                                            expected['state_type'])
            self.assertEqual(r.state, expected['state'])
            self.assertEqual(r.state_type, expected['state_type'])
            self.probe.submit_result(r)

    def test_expiration(self):
        expiration = 30
        now = time.time()
        t = self.probe.get_task()
        t.created = now
        self.assertFalse(self.probe.expire_task(t, expiration))
        t.created = now - (expiration + 5)
        self.assertTrue(self.probe.expire_task(t, expiration))



import glob
import os
import logging
import hashlib

logger = logging.getLogger(__name__)

import yaml

from nymms import exceptions


class EmptyConfig(exceptions.NymmsException):
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return "Config file %s resulted in an empty config." % (self.filename)


def open_config_file(config_file):
    """ Opens a config file, logging common IOError exceptions.
    """
    try:
        return open(config_file)
    except IOError, e:
        # This should only happen with the top level config file, since
        # we use glob.glob on includes
        if e.errno == 2:
            logger.error("Could not find file '%s'.", config_file)
        elif e.errno == 13:
            logger.error("Invalid permissions to open '%s'.", config_file)
        raise


def load_config(config_file):
    stack = []
    root = os.path.split(os.path.abspath(os.path.expanduser(config_file)))[0]

    def recursive_preprocess(filename):
        filename = os.path.expanduser(filename)
        stack.append(os.path.abspath(filename))
        c = []

        with open_config_file(filename) as fd:
            for lineno, line in enumerate(fd):
                line = line.rstrip()
                if line.startswith('!include '):
                    path = line.split(None, 1)[1]
                    # if the include doesn't have a fully qualified path then
                    # assume the relative path is based off the directory of
                    # the initial config file
                    if not path.startswith('/'):
                        path = os.path.join(root, path)
                    files = glob.glob(path)
                    if not files:
                        logger.warning("Include statement '%s' at %s:%d did "
                                       "not match any files. Skipping.", line,
                                       filename, lineno)
                        continue
                    for f in files:
                        f = os.path.abspath(f)
                        if f in stack:
                            logger.warning("Already parsed %s, skipping "
                                           "(%s:%d) to avoid infinite loop.",
                                           f, filename, lineno)
                            continue
                        if os.path.isfile(f):
                            logger.debug("Parsing include (%s:%d): %s",
                                         filename, lineno, f)
                            c.extend(recursive_preprocess(f))
                        else:
                            logger.warning("%s is not a regular file, "
                                           "skipping (%s:%d).", f, filename,
                                           lineno)
                    continue
                c.append(line)
        return c
    logger.debug("Loading config file: %s", config_file)
    config = recursive_preprocess(config_file)
    if not config:
        raise EmptyConfig(config_file)
    config = os.linesep.join(config)
    version = hashlib.sha512(config).hexdigest()
    return (version, yaml.safe_load(config))



from jinja2 import Template, StrictUndefined


class NymmsTemplate(Template):
    """ Throws exceptions for missing context values. """
    def __new__(cls, *args, **kwargs):
        kwargs['undefined'] = StrictUndefined
        return super(NymmsTemplate, cls).__new__(cls, *args, **kwargs)



import boto
import imp

CONNECT_MAP = {
    'beanstalk': ['beanstalk'],
    'cloudformation': ['cloudformation'],
    'cloudsearch': ['cloudsearch'],
    'dynamodb': ['dynamodb'],
    'dynamodb2': ['dynamodb2'],
    'ec2': ['ec2'],
    'autoscale': ['ec2', 'autoscale'],
    'cloudwatch': ['ec2', 'cloudwatch'],
    'elb': ['ec2', 'elb'],
    'elasticache': ['elasticache'],
    'elastictranscoder': ['elastictranscoder'],
    'emr': ['emr'],
    'glacier': ['glacier'],
    'iam': ['iam'],
    'rds': ['rds'],
    'redshift': ['redshift'],
    'route53': ['route53'],
    's3': ['s3'],
    'sdb': ['sdb'],
    'ses': ['ses'],
    'sns': ['sns'],
    'sqs': ['sqs'],
    'sts': ['sts'],
    'support': ['support'],
    'swf': ['swf'],
    'vpc': ['vpc'],
}


class ConnectionManager(object):
    """ Used to setup and maintain AWS service connections in a single region.

    This acts as a proxy for AWS connections to all AWS services that boto
    provides a connect_to_region method for.
    """
    def __init__(self, region='us-east-1', **kw_params):
        self.region = region
        self.params = kw_params

    def __getattr__(self, attr):
        try:
            modules = CONNECT_MAP[attr]
        except KeyError:
            raise AttributeError(attr)

        current_mod = boto
        for m in modules:
            x = imp.find_module(m, current_mod.__path__)
            current_mod = imp.load_module(attr, *x)
        connect_to_region = getattr(current_mod, 'connect_to_region')
        conn = connect_to_region(self.region, **self.params)
        setattr(self, attr, conn)
        return conn


class SNSTopic(object):
    def __init__(self, conn_mgr, topic_name):
        self._conn = conn_mgr.sns
        self._topic_name = topic_name
        self._topic_arn = None

    def _setup_topic(self):
        if self._topic_arn:
            return

        conn = self._conn
        response = conn.create_topic(self._topic_name)['CreateTopicResponse']
        self._topic_arn = response['CreateTopicResult']['TopicArn']

    def publish(self, *args, **kwargs):
        self._setup_topic()
        return self._conn.publish(self._topic_arn, *args, **kwargs)

    def subscribe_sqs_queue(self, *args, **kwargs):
        self._setup_topic()
        return self._conn.subscribe_sqs_queue(self._topic_arn, *args, **kwargs)



import logging
import time
import importlib
import sys

logger = logging.getLogger(__name__)


def retry_on_exception(exception_list, retries=3, reset_func=None,
                       final_exception=None, delay=0):
    """ A decorator that executes a function and catches any exceptions in
    'exception_list'.  It then retries 'retries' with 'delay' seconds between
    retries and executing 'reset_func' each time.  If it fails after reaching
    the retry limit it then raises 'final_exception' or the last exception
    raised.
    """
    def decorator(func):
        def wrapped(*args, **kwargs):
            i = 0
            while True:
                try:
                    return func(*args, **kwargs)
                except exception_list as e:
                    if reset_func:
                        reset_func()
                    if delay:
                        if callable(delay):
                            time.sleep(delay(i))
                        else:
                            time.sleep(delay)
                    logger.warn("%s exception caught.  Retrying %d time(s): "
                                "%s", e.__class__.__name__, retries - i,
                                e.message)
                i += 1
                if retries and i > retries:
                    break
            if final_exception:
                raise final_exception(str(e))
            else:
                raise e
        return wrapped
    return decorator


def load_object_from_string(fqcn):
    """ Given a '.' delimited string representing the full path to an object
    (function, class, variable) inside a module, return that object.  Example:

    load_object_from_string('os.path.basename')
    load_object_from_string('logging.Logger')
    etc
    """
    module_parts = fqcn.split('.')
    object_name = module_parts[-1]
    path_parts = module_parts[:-1]
    module = importlib.import_module('.'.join(path_parts))
    return getattr(module, object_name)



import logging
from logging.handlers import TimedRotatingFileHandler

logger = logging.getLogger(__name__)

from nymms.reactor.handlers.Handler import Handler


class LogHandler(Handler):
    """ A basic handler to send alerts to a log file via python's logging
    module.
    """
    def _setup_logger(self):
        if getattr(self, '_file_logger', None):
            return
        filename = self.config['filename']
        when = self.config['when']
        interval = self.config['interval']
        backup_count = self.config['backup_count']
        handler = TimedRotatingFileHandler(filename, when, interval,
                                           backup_count)
        handler.setLevel(logging.INFO)
        msg_fmt = '[%(asctime)s] %(message)s'
        handler.setFormatter(logging.Formatter(msg_fmt))
        self._file_logger = logging.getLogger('LogHandler')
        self._file_logger.propagate = False
        self._file_logger.addHandler(handler)
        self._file_logger.setLevel(logging.INFO)

    def process(self, result, previous_state):
        self._setup_logger()
        self._file_logger.info(result.serialize())



import unittest

from nymms import results
from nymms.data_types import ValidationError, MissingRequiredField


class TestResult(unittest.TestCase):
    def setUp(self):
        self.result = results.Result(
            'www1:check_http',
            state=results.OK,
            state_type=results.HARD,
            timestamp=123456,
            output='OK: check_http is ok',
            task_context={'url': 'http://example.com/'})

    def test_invalid_state_name(self):
        self.result.state = 'foo'
        with self.assertRaises(ValidationError):
            self.result.serialize()

    def test_invalid_state_type_name(self):
        self.result.state_type = 'foo'
        with self.assertRaises(ValidationError):
            self.result.validate()

    def test_missing_state(self):
        delattr(self.result, 'state')
        with self.assertRaises(MissingRequiredField):
            self.result.validate()

    def test_serialize_deserialize(self):
        serialize1 = self.result.serialize()
        new_result = results.Result.deserialize(serialize1)
        serialize2 = new_result.serialize()
        self.assertEqual(serialize1, serialize2)



import subprocess
import signal
import logging

logger = logging.getLogger(__name__)


class CommandException(Exception):
    pass


class CommandTimeout(CommandException):
    def __init__(self, command, timeout):
        self.command = command
        self.timeout = timeout

    def __str__(self):
        return "Command '%s' took longer than %d seconds to execute." % (
            self.command, self.timeout)


class CommandFailure(CommandException):
    def __init__(self, command, return_code, output):
        self.command = command
        self.return_code = return_code
        self.output = output

    def __str__(self):
        return "Command '%s' exited with a return code of %d." % (
            self.command, self.return_code,)


def execute(command_string, timeout=None):
    """
    Execute a command with an optional timeout.  If the command takes longer
    than timeout raise a CommandTimeout exception.  If the command fails raise
    a CommandFailure exception.  Otherwise return stdout & stderr from the
    command.
    """
    def handle_sigalrm(signum, frame):
        if signum == signal.SIGALRM:
            logger.debug("Command '%s' timed out after %d seconds." % (
                command_string, timeout))
            raise CommandTimeout(command_string, timeout)
    signal.signal(signal.SIGALRM, handle_sigalrm)
    log_header = "Executing command:"
    # If a timeout is given, lets setup an alarm signal
    if timeout:
        log_header += " (timeout: %d)" % (timeout)
        signal.alarm(timeout)
    # Execute the command
    logger.debug(log_header)
    logger.debug("    %s" % (command_string))
    command_object = subprocess.Popen(command_string, shell=True,
                                      stdout=subprocess.PIPE,
                                      stderr=subprocess.STDOUT)
    try:
        output = command_object.communicate()
    except CommandTimeout:
        logger.debug("Command timed out, terminating child command.")
        command_object.terminate()
        raise
    if not command_object.returncode == 0:
        signal.alarm(0)
        logger.debug("Command '%s' failed with return code %d:" % (
            command_string, command_object.returncode))
        for line in output[0].split('\n'):
            logger.debug("    output: %s" % (line))
        raise CommandFailure(command_string, command_object.returncode,
                             output[0])
    signal.alarm(0)
    return output[0]



import logging
import time

logger = logging.getLogger(__name__)

from nymms.data_types import NymmsDataType


class Task(NymmsDataType):
    def __init__(self, object_id, created=None, attempt=0, context=None,
                 origin=None):
        super(Task, self).__init__(object_id=object_id, origin=origin)
        self.attempt = attempt or 0
        self.created = created or time.time()
        self.context = context

    def validate_created(self):
        self.created = int(self.created)

    def validate_attempt(self):
        self.attempt = int(self.attempt)

    def increment_attempt(self):
        self.attempt += 1



import unittest

from nymms import registry
from nymms.resources import Command, MonitoringGroup
from weakref import WeakValueDictionary


class TestRegistry(unittest.TestCase):
    def tearDown(self):
        # Ensure we have a fresh registry after every test
        Command.registry.clear()

    def test_empty_registry(self):
        self.assertEqual(Command.registry, WeakValueDictionary())

    def test_register_object(self):
        # First test it's empty
        self.assertEqual(Command.registry, WeakValueDictionary())
        # Add a command
        command = Command('test_command', '/bin/true')
        # verify that there is only a single command in the registry
        self.assertEqual(len(Command.registry), 1)
        # Verify that the registered command is the same as command
        self.assertIs(Command.registry[command.name], command)

    def test_duplicate_register(self):
        # add a command
        Command('test_command', '/bin/true')
        with self.assertRaises(registry.DuplicateEntryError):
            Command('test_command', '/bin/true')

    def test_invalid_resource_register(self):
        with self.assertRaises(TypeError):
            Command.registry['test'] = MonitoringGroup('test_group')



import logging
import time
import json

from boto.sqs.message import Message

from nymms import resources
from nymms.config import config
from nymms.tasks import Task
from nymms.scheduler.Scheduler import Scheduler

logger = logging.getLogger(__name__)


class AWSScheduler(Scheduler):
    def __init__(self, node_backend, conn_mgr, task_queue):
        self._node_backend = node_backend
        self._conn = conn_mgr
        self._queue_name = task_queue
        self._default_queue = None
        self._realm_queues = {}
        super(Scheduler, self).__init__()

    def _set_expiration(self, queue, expiration):
        if expiration:
            logger.debug("Setting queue %s message retention to %d.",
                         queue.name, expiration)
            queue.set_attribute('MessageRetentionPeriod',
                                expiration)

    def _setup_realm(self, realm, **kwargs):
        if realm in self._realm_queues:
            return
        queue_name = self._queue_name + '_REALM_' + realm
        logger.debug("setting up realm queue %s", queue_name)
        queue = self._conn.sqs.create_queue(queue_name)
        self._set_expiration(queue, kwargs.get('task_expiration', None))
        self._realm_queues[realm] = queue

    def _setup_queue(self, **kwargs):
        if self._default_queue:
            return
        logger.debug("setting up queue %s", self._queue_name)
        queue = self._conn.sqs.create_queue(self._queue_name)
        self._set_expiration(queue, kwargs.get('task_expiration', None))
        self._default_queue = queue

    def _choose_queue(self, task, **kwargs):
        realm = task.context['realm']
        if realm:
            self._setup_realm(realm, **kwargs)
            queue = self._realm_queues[realm]
        else:
            self._setup_queue(**kwargs)
            queue = self._default_queue
        return queue

    def submit_task(self, task, **kwargs):
        queue = self._choose_queue(task, **kwargs)
        logger.debug("Sending task '%s' to queue '%s'.", task.id,
                     queue.name)
        m = Message()
        m.set_body(json.dumps(task.serialize()))
        return queue.write(m)



import logging

from nymms.scheduler.backends.Backend import Backend
from nymms.config import yaml_config

logger = logging.getLogger(__name__)


class YamlBackend(Backend):
    def __init__(self, path):
        self.path = path

    def load_nodes(self):
        version, nodes = yaml_config.load_config(self.path)
        logger.debug("Loaded node config (%s) from %s.", version, self.path)
        return nodes



import logging

logger = logging.getLogger(__name__)

from nymms import results


def always_true(result, previous_state):
    """ Not really necessary since no filters results in an always true result,
    but this is useful to show an example of what a filter is without actually
    doing anything.
    """
    return True


def hard_state(result, previous_state):
    if result.state_type == results.HARD:
        logger.debug("%s state_type is HARD.", result.id)
        return True
    return False


def changed_state(result, previous_state):
    """ Only alert if the state is new or has either changed state or
    state_type.
    """
    if not previous_state:
        logger.debug("No previous state found.")
        return True
    if not previous_state.state == result.state:
        logger.debug("Previous state (%s) does not match current "
                     "state (%s).", previous_state.state_name,
                     result.state_name)
        return True
    if not previous_state.state_type == result.state_type:
        logger.debug("Previous state_type (%s) does not match current "
                     "state_type (%s).",
                     previous_state.state_type_name,
                     result.state_type_name)
        return True
    return False


def ok_state(result, previous_state):
    if result.state == results.OK:
        return True
    return False


def warning_state(result, previous_state):
    if result.state == results.WARNING:
        return True
    return False


def critical_state(result, previous_state):
    if result.state == results.CRITICAL:
        return True
    return False


def unknown_state(result, previous_state):
    if result.state >= results.UNKNOWN:
        return True
    return False


def not_ok_state(result, previous_state):
    return not(ok_state(result, previous_state))



import logging
import time

logger = logging.getLogger(__name__)

import nymms
from nymms import results
from nymms.daemon import NymmsDaemon
from nymms.resources import Monitor
from nymms.utils import commands
from nymms.config.yaml_config import load_config, EmptyConfig
from nymms.exceptions import MissingCommandContext


class Probe(NymmsDaemon):
    def get_private_context(self, private_context_file):
        if not private_context_file:
            return None
        try:
            return load_config(private_context_file)[1]
        except (IOError, EmptyConfig) as e:
            logger.exception("Unable to open private context file: %s",
                             private_context_file)
            return None

    # TODO: This calls on _state_backend but setting up of the _state_backend
    #       needs to be handled in the subclass.  Not sure how I should handle
    #       this, but I really like the idea of these being base class
    #       methods since in reality all reactors should have some sort of
    #       state backend, even if its a no-op
    def get_state(self, task_id):
        return self._state_backend.get_state(task_id)

    def get_task(self, **kwargs):
        raise NotImplementedError

    def resubmit_task(self, task, delay, **kwargs):
        raise NotImplementedError

    def submit_result(self, result, **kwargs):
        raise NotImplementedError

    def execute_task(self, task, timeout, **kwargs):
        log_prefix = "%s - " % (task.id,)
        monitor = Monitor.registry[task.context['monitor']['name']]
        command = monitor.command.command_string
        current_attempt = int(task.attempt) + 1
        logger.debug(log_prefix + "attempt %d, executing: %s", current_attempt,
                     command)
        result = results.Result(task.id, timestamp=task.created,
                                task_context=task.context)
        try:
            output = monitor.execute(task.context, timeout,
                                     self._private_context)
            result.output = output
            result.state = results.OK
        except commands.CommandException as e:
            if isinstance(e, commands.CommandFailure):
                result.state = e.return_code
                result.output = e.output
            if isinstance(e, commands.CommandTimeout):
                result.state = results.UNKNOWN
                result.output = ("Command timed out after %d seconds." % 
                                 timeout)
        except Exception as e:
            result.state = results.UNKNOWN
            result.output = str(e)
        return result

    def expire_task(self, task, task_expiration):
        if task_expiration:
            now = time.time()
            task_lifetime = now - task.created
            if task_lifetime > task_expiration:
                logger.debug("Task %s is older than expiration limit %d. "
                             "Skipping." % (task.id, task_expiration))
                return True
            return False
        return False

    def handle_task(self, task, **kwargs):
        log_prefix = "%s - " % (task.id,)
        task_expiration = kwargs.get('task_expiration', None)
        if self.expire_task(task, task_expiration):
            return None
        previous_state = self.get_state(task.id)
        # check if the timeout is defined on the task first, if not then
        # go with what was passed into handle_task via run
        timeout = task.context.get('monitor_timeout',
                                   kwargs.get('monitor_timeout'))
        max_retries = task.context.get('max_retries',
                                       kwargs.get('max_retries'))
        last_attempt = int(task.attempt)
        current_attempt = last_attempt + 1
        result = self.execute_task(task, timeout, **kwargs)
        result.state_type = results.HARD
        # Trying to emulate this:
        # http://nagios.sourceforge.net/docs/3_0/statetypes.html
        if result.state == results.OK:
            if (previous_state and not previous_state.state == results.OK and
                previous_state.state_type == results.SOFT):
                    result.state_type = results.SOFT
        else:
            logger.debug(log_prefix + "current_attempt: %d, max_retries: %d", 
                         current_attempt, max_retries)
            if current_attempt <= max_retries:
                # XXX Hate this logic - hope to find a cleaner way to handle
                #     it someday.
                if (not previous_state or
                        previous_state.state_type == results.SOFT or
                        previous_state.state == results.OK):
                    result.state_type = results.SOFT
                    delay = task.context.get('retry_delay',
                                             kwargs.get('retry_delay'))
                    delay = max(delay, 0)
                    logger.debug('Resubmitting task with %ds delay.' % delay)
                    self.resubmit_task(task, delay, **kwargs)
            else:
                logger.debug("Retry limit hit, not resubmitting.")
        return result

    def run(self, **kwargs):
        """ This will run in a tight loop. It is expected that the subclass's
        get_task() method will introduce a delay if the results queue is
        empty.
        """
        private_context_file = kwargs.get('private_context_file', None)
        self._private_context = self.get_private_context(private_context_file)
        while True:
            task = self.get_task(**kwargs)
            if not task:
                logger.debug("Task queue is empty.")
                continue
            result = self.handle_task(task, **kwargs)
            if result:
                self.submit_result(result, **kwargs)
            task.delete()



import argparse
import logging

logger = logging.getLogger(__name__)

from nymms.utils import logutil


def setup_logging(verbose=False):
    level = logutil.INFO
    if verbose:
        level = logutil.DEBUG
    if not verbose > 1:
        logutil.quiet_boto_logging()
    return logutil.setup_root_logger(stdout=level)


class NymmsDaemonCommand(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super(NymmsDaemonCommand, self).__init__(*args, **kwargs)
        self.add_argument('-v', '--verbose', action='count', default=0)
        self.add_argument('-c', '--config', default='/etc/nymms/config.yaml')



import logging

logger = logging.getLogger(__name__)

from nymms import results
from nymms.exceptions import OutOfDateState


class SDBStateBackend(object):
    def __init__(self, conn, domain_name):
        self._conn = conn
        self._domain_name = domain_name
        self._domain = None
        logger.debug("%s initialized.", self.__class__.__name__)

    def _setup_domain(self):
        if self._domain:
            return
        conn = self._conn
        logger.debug("setting up state domain %s", self._domain_name)
        self._domain = conn.create_domain(self._domain_name)

    def _build_new_state(self, task_id, result, previous):
        new_state = results.StateRecord(task_id,
                                        state=result.state,
                                        state_type=result.state_type)
        # Only update last_state_change if the state has changed to a new
        # HARD state_type state, otherwise we use the previous
        # last_state_change
        if previous:
            if (new_state.state_type == results.SOFT or
                    previous.state == new_state.state):
                new_state.last_state_change = previous.last_state_change

        new_state.validate()
        return new_state

    def save_state(self, task_id, result, previous):
        self._setup_domain()
        new_state = self._build_new_state(task_id, result, previous)
        expected_value = ['last_update', False]
        if previous:
            expected_value = ['last_update', previous.last_update]
            if previous.last_update > new_state.last_update:
                logger.warning(task_id + " - found previous state that is "
                               "newer than current state.  Discarding.")
                logger.warning(task_id + " - previous state: %s",
                               previous.serialize())
                logger.warning(task_id + " - current state: %s",
                               new_state.serialize())
                raise OutOfDateState(new_state, previous)
        logger.debug(task_id + " - saving state: %s", new_state.serialize())
        self._domain.put_attributes(task_id, new_state.serialize(),
                                    replace=True,
                                    expected_value=expected_value)

    def get_state(self, task_id):
        self._setup_domain()
        logger.debug("%s - getting state", task_id)
        state_item = self._domain.get_item(task_id, consistent_read=True)
        state = None
        if state_item:
            state = results.StateRecord.deserialize(state_item)
            state.validate()
        else:
            logger.debug("%s - no state found", task_id)
        return state



import unittest

from nymms import results
from nymms.reactor import filters

class TestFilters(unittest.TestCase):
    def setUp(self):
        self.result = results.Result('test:filter', state=results.OK,
                                     state_type=results.HARD)
        self.result.validate()
        self.record = results.StateRecord('test:filter', state=results.OK,
                                          state_type=results.HARD)
        self.record.validate()

    def test_hard_state(self):
        self.assertTrue(filters.hard_state(self.result, self.record))

        self.result.state_type = results.SOFT
        self.result.validate()
        self.assertFalse(filters.hard_state(self.result, self.record))

    def test_ok_state(self):
        self.assertTrue(filters.ok_state(self.result, self.record))

        self.result.state = results.WARNING
        self.result.validate()
        self.assertFalse(filters.ok_state(self.result, self.record))

    def test_not_ok_state(self):
        self.assertFalse(filters.not_ok_state(self.result, self.record))

        self.result.state = results.WARNING
        self.result.validate()
        self.assertTrue(filters.not_ok_state(self.result, self.record))

    def test_warning_state(self):
        self.assertFalse(filters.warning_state(self.result, self.record))

        self.result.state = results.WARNING
        self.result.validate()
        self.assertTrue(filters.warning_state(self.result, self.record))

    def test_critical_state(self):
        self.assertFalse(filters.critical_state(self.result, self.record))

        self.result.state = results.CRITICAL
        self.result.validate()
        self.assertTrue(filters.critical_state(self.result, self.record))

    def test_unknown_state(self):
        self.assertFalse(filters.unknown_state(self.result, self.record))

        self.result.state = results.UNKNOWN
        self.result.validate()
        self.assertTrue(filters.unknown_state(self.result, self.record))

    def test_changed_state(self):
        f = filters.changed_state
        self.assertFalse(f(self.result, self.record))

        self.assertTrue(f(self.result, None))

        self.result.state = results.CRITICAL
        self.result.validate()
        self.assertTrue(f(self.result, self.record))

        self.result.state = results.OK
        self.result.state_type = results.SOFT
        self.result.validate()
        self.assertTrue(f(self.result, self.record))



import unittest

from nymms import utils


class TestBase(unittest.TestCase):
    def test_load_class_from_string(self):
        from logging import Logger
        x = utils.load_object_from_string('logging.Logger')
        self.assertEqual(x, Logger)



import os
import sys
from setuptools import setup, find_packages
import glob

src_dir = os.path.dirname(__file__)

def get_version(src_dir):
    sys.path.insert(0, src_dir)
    import nymms
    return nymms.__version__


def read(filename):
    full_path = os.path.join(src_dir, filename)
    with open(full_path) as fd:
        return fd.read()


if __name__ == '__main__':
    setup(
        name='nymms',
        version=get_version(src_dir),
        author='Michael Barrett',
        author_email='loki77@gmail.com',
        license="New BSD license",
        url="https://github.com/cloudtools/nymms",
        description='Not Your Mother\'s Monitoring System (NYMMS)',
        long_description=read('README.rst'),
        classifiers=[
            "Topic :: System :: Monitoring",
            "License :: OSI Approved :: BSD License",
            "Development Status :: 3 - Alpha"],
        packages=find_packages(),
        scripts=glob.glob(os.path.join(src_dir, 'scripts', 'nymms_*')),
    )



import unittest

from jinja2.exceptions import UndefinedError

from nymms.utils import templates


class TestNymmsTemplate(unittest.TestCase):
    def test_render_missing(self):
        s = "{{ found_key }} {{ missing_key }}"
        context = {'found_key': 'FOUND'}
        with self.assertRaises(UndefinedError):
            t = templates.NymmsTemplate(s)
            t.render(context)



import logging

from nymms import exceptions

logger = logging.getLogger(__name__)


class DuplicateEntryError(exceptions.NymmsException):
    def __init__(self, name, obj, registry):
        self.name = name
        self.obj = obj
        self.registry = registry

    def __str__(self):
        return "Duplicate entry in '%s' registry for '%s'." % (
            self.registry._object_type.__name__, self.name)


class Registry(dict):
    def __init__(self, object_type, *args, **kwargs):
        logging.debug("New '%s' registry id: %d" % (object_type, id(self)))
        self._object_type = object_type
        dict.__init__(self, *args, **kwargs)

    def __setitem__(self, name, value):
        if not isinstance(value, self._object_type):
            raise TypeError("This registry only accepts objects of "
                            "type %s." % (self._object_type.__name__))

        if name in self:
            raise DuplicateEntryError(name, value, self)
        dict.__setitem__(self, name, value)



import unittest
import os

from nymms.config import yaml_config


class TestIncludeLoader(unittest.TestCase):
    def setUp(self):
        self.root = os.path.split(__file__)[0]

    def test_relative_include(self):
        full_path = os.path.join(self.root, 'config.yaml')
        version, relative_config = yaml_config.load_config(full_path)
        self.assertEqual(relative_config['foo'], 'bar')
        self.assertEqual(relative_config['file1'], 1)

    def test_missing_config(self):
        with self.assertRaises(IOError):
            yaml_config.load_config('nonexistant.yaml')



import logging
import time

import nymms
from nymms.daemon import NymmsDaemon
from nymms.resources import Node
from nymms.tasks import Task

logger = logging.getLogger(__name__)


class Scheduler(NymmsDaemon):
    task_id_template = "{node[name]}:{monitor[name]}"

    def __init__(self, node_backend):
        self._node_backend = node_backend
        super(Scheduler, self).__init__()

    def get_tasks(self):
        tasks = {}
        self._node_backend._load_nodes()
        nodes = Node.registry
        for node_name, node in nodes.iteritems():
            tasks[node_name] = node.monitors
        return tasks

    def submit_task(self, task, **kwargs):
        raise NotImplementedError

    def run(self, **kwargs):
        while True:
            start = time.time()
            tasks = self.get_tasks()
            # This is done to make sure we submit one task per node until we've
            # submitted all the tasks.  This helps ensure we don't hammer a
            # single node with monitoring tasks
            while True:
                working_index = tasks.keys()
                if not working_index:
                    break
                for node in working_index:
                    try:
                        task_context = tasks[node].pop()
                        task_id = self.task_id_template.format(**task_context)
                        task = Task(task_id, context=task_context)
                    except IndexError:
                        del(tasks[node])
                        continue
                    self.submit_task(task, **kwargs)
            run_time = time.time() - start
            logger.debug("Scheduler iteration took %d seconds.", run_time)
            interval = kwargs.get('interval')
            sleep_time = interval - max(run_time, 0)
            logger.debug("Scheduler sleeping for %d seconds.", sleep_time)
            time.sleep(sleep_time)



import logging

logger = logging.getLogger(__name__)

from nymms.reactor.handlers.Handler import Handler
from nymms import results

from jinja2 import Template

try:
    import pagerduty
except ImportError:
    logger.error("Unable to import the pagerduty module.")
    logger.error("Please install it from here: ")
    logger.error("  https://pypi.python.org/pypi/pagerduty/")
    logger.error("(You can use pip, ie: pip install pagerduty)")
    raise


MISSING_SUBJECT = 'Handler %s missing subject_template.'


class PagerDutyHandler(Handler):
    """ A basic handler to send alerts to people via email through Amazon's
    SES service.  Sends every result it receives by default.  To filter
    results you should subclass this and provide a _filter method.

    config options:
      enabled: bool
      subject_template: string, jinja2 Template string
      service_keys: list(string), pagerduty service keys
      filters: list(string), filters
    """
    def _connect(self):
        if getattr(self, '_endpoints', None):
            return
        self._endpoints = []
        service_keys = self.config.get('service_keys', [])
        if not service_keys:
            logger.warning("No service_keys configured for Handler %s.",
                           self.__class__.__name__)
            return
        for key in service_keys:
            logger.debug("Initializing pagerduty service endpoint %s.",
                         key)
            self._endpoints.append(pagerduty.PagerDuty(key))

    def _send_incident(self, result, previous_state):
        self._connect()
        subject_template = self.config.get('subject_template',
           MISSING_SUBJECT % (self.__class__.__name__))
        description = Template(subject_template)
        result_data = result.serialize()
        for ep in self._endpoints:
            logger.debug("Submitting to pagerduty service_key %s.",
                         ep.service_key)
            ep.trigger(description=description.render(result_data),
                       incident_key=result.id,
                       details=result_data)

    def process(self, result, previous_state):
        self._send_incident(result, previous_state)



import unittest
from weakref import WeakValueDictionary

from nymms import resources
from nymms.exceptions import MissingCommandContext


class TestNanoResources(unittest.TestCase):
    def test_reserved_attributes(self):
        with self.assertRaises(TypeError):
            resources.Command(name='test', command_string='test',
                              address='10.0.0.1')

    def test_resource_context(self):
        c = resources.Command(name='test', command_string='test')
        context = c._context()
        self.assertEqual(context.keys()[0], 'command')
        self.assertEqual(context['command']['name'], c.name)

    def test_extra_attributes(self):
        extra_attribute_name = 'extra'
        extra2_value = 'extra2'
        c1 = resources.Command(name='test1', command_string='test1')
        c2 = resources.Command(name='test2', command_string='test2',
                               extra=extra2_value)
        with self.assertRaises(KeyError):
            c1.extra_attributes[extra_attribute_name]
        self.assertEquals(c2.extra_attributes[extra_attribute_name],
                          extra2_value)


class TestNode(unittest.TestCase):
    def test_adding_monitoring_groups(self):
        mg1 = resources.MonitoringGroup('mg1')
        self.assertEqual(mg1.nodes, WeakValueDictionary())
        mg2 = resources.MonitoringGroup('mg2')
        self.assertEqual(mg2.nodes, WeakValueDictionary())
        node = resources.Node(name='node1', address='127.0.0.1',
                              monitoring_groups=[mg1, mg2])
        self.assertIn(node, mg1.nodes.values())
        self.assertIn(node, mg2.nodes.values())


class TestCommand(unittest.TestCase):
    def test_format_command(self):
        command = "/bin/echo {{public}} {{__private.password}}"
        context = {'public': 'public'}
        private_context = {'password': 'mypassword'}
        c = resources.Command('echo', command)
        with self.assertRaises(MissingCommandContext):
            c.format_command(context)
        c_out = c.format_command(context, private_context)
        self.assertEquals(c_out,
                          "/bin/echo public mypassword")



import logging
import json

logger = logging.getLogger(__name__)

from nymms import results
from nymms.reactor.Reactor import Reactor
from nymms.utils.aws_helper import SNSTopic
from nymms.state.sdb_state import SDBStateBackend

from boto.sqs.message import RawMessage


class AWSReactor(Reactor):
    def __init__(self, conn_mgr, topic_name, state_domain_name, queue_name,
                 state_backend=SDBStateBackend):
        self._conn = conn_mgr
        self._topic_name = topic_name
        self._queue_name = queue_name
        self._topic = None
        self._queue = None
        self._state_backend = state_backend(conn_mgr.sdb, state_domain_name)
        super(AWSReactor, self).__init__()

    def _setup_queue(self):
        if self._queue:
            return
        logger.debug("setting up queue %s", self._queue_name)
        self._queue = self._conn.sqs.create_queue(self._queue_name)
        self._queue.set_message_class(RawMessage)

    def _setup_topic(self):
        if self._topic:
            return
        logger.debug("setting up topic %s", self._topic_name)
        self._topic = SNSTopic(self._conn, self._topic_name)
        logger.debug("subscribing queue %s to topic %s", self._queue_name,
                     self._topic_name)
        self._topic.subscribe_sqs_queue(self._queue)

    def get_result(self, **kwargs):
        wait_time = kwargs.get('wait_time', 0)
        visibility_timeout = kwargs.get('visibility_timeout', None)
        self._setup_queue()
        self._setup_topic()
        logger.debug("Getting result from queue %s.", self._queue_name)
        result = self._queue.read(visibility_timeout=visibility_timeout,
                                  wait_time_seconds=wait_time)
        result_obj = None
        if result:
            result_message = json.loads(result.get_body())['Message']
            result_dict = json.loads(result_message)
            result_obj = results.Result.deserialize(result_dict,
                                                    origin=result)
            result_obj.validate()
        return result_obj



import unittest

from nymms.utils import commands


class TestCommands(unittest.TestCase):
    def test_execute_failure(self):
        with self.assertRaises(commands.CommandFailure):
            # Non-existant command
            out = commands.execute('xxxps auwwwx', 10)

    def test_execute_timeout(self):
        with self.assertRaises(commands.CommandTimeout):
            out = commands.execute('sleep 2', 1)

    def test_execute_success(self):
        out = commands.execute('echo test', 10)
        self.assertEqual(out, 'test\n')
